@@ -1,6 +1,7 @@
 import mimetypes
 import os
 import time
+import types
 import warnings
 from typing import (
     overload,
@@ -13,43 +14,45 @@ from typing import (
     Dict,
 )
 from urllib.parse import urlparse
+from functools import partial
+from docarray import DocumentArray
 
 if TYPE_CHECKING:
     import numpy as np
-    from docarray import DocumentArray, Document
+    from docarray import Document
 
 
 class Client:
-    def __init__(self, server: str):
+    def __init__(self, server: str, credential: dict = {}, **kwargs):
         """Create a Clip client object that connects to the Clip server.
-
         Server scheme is in the format of `scheme://netloc:port`, where
             - scheme: one of grpc, websocket, http, grpcs, websockets, https
             - netloc: the server ip address or hostname
             - port: the public port of the server
-
         :param server: the server URI
+        :param credential: the credential for authentication {'Authentication': '<token>'}
         """
         try:
             r = urlparse(server)
             _port = r.port
-            _scheme = r.scheme
-            if not _scheme:
-                raise
+            self._scheme = r.scheme
         except:
             raise ValueError(f'{server} is not a valid scheme')
 
         _tls = False
-
-        if _scheme in ('grpcs', 'https', 'wss'):
-            _scheme = _scheme[:-1]
+        if self._scheme in ('grpcs', 'https', 'wss'):
+            self._scheme = self._scheme[:-1]
             _tls = True
 
-        if _scheme == 'ws':
-            _scheme = 'websocket'  # temp fix for the core
+        if self._scheme == 'ws':
+            self._scheme = 'websocket'  # temp fix for the core
+            if credential:
+                warnings.warn(
+                    'Credential is not supported for websocket, please use grpc or http'
+                )
 
-        if _scheme in ('grpc', 'http', 'websocket'):
-            _kwargs = dict(host=r.hostname, port=_port, protocol=_scheme, tls=_tls)
+        if self._scheme in ('grpc', 'http', 'websocket'):
+            _kwargs = dict(host=r.hostname, port=_port, protocol=self._scheme, tls=_tls)
 
             from jina import Client
 
@@ -58,6 +61,8 @@ class Client:
         else:
             raise ValueError(f'{server} is not a valid scheme')
 
+        self._authorization = credential.get('Authorization', None)
+
     @overload
     def encode(
         self,
@@ -65,19 +70,18 @@ class Client:
         *,
         batch_size: Optional[int] = None,
         show_progress: bool = False,
+        parameters: Optional[dict] = None,
     ) -> 'np.ndarray':
         """Encode images and texts into embeddings where the input is an iterable of raw strings.
-
         Each image and text must be represented as a string. The following strings are acceptable:
-
             - local image filepath, will be considered as an image
             - remote image http/https, will be considered as an image
             - a dataURI, will be considered as an image
             - plain text, will be considered as a sentence
-
         :param content: an iterator of image URIs or sentences, each element is an image or a text sentence as a string.
         :param batch_size: the number of elements in each request when sending ``content``
         :param show_progress: if set, show a progress bar
+        :param parameters: the parameters for the encoding, you can specify the model to use when you have multiple models
         :return: the embedding in a numpy ndarray with shape ``[N, D]``. ``N`` is in the same length of ``content``
         """
         ...
@@ -89,12 +93,13 @@ class Client:
         *,
         batch_size: Optional[int] = None,
         show_progress: bool = False,
+        parameters: Optional[dict] = None,
     ) -> 'DocumentArray':
         """Encode images and texts into embeddings where the input is an iterable of :class:`docarray.Document`.
-
         :param content: an iterable of :class:`docarray.Document`, each Document must be filled with `.uri`, `.text` or `.blob`.
         :param batch_size: the number of elements in each request when sending ``content``
         :param show_progress: if set, show a progress bar
+        :param parameters: the parameters for the encoding, you can specify the model to use when you have multiple models
         :return: the embedding in a numpy ndarray with shape ``[N, D]``. ``N`` is in the same length of ``content``
         """
         ...
@@ -109,19 +114,26 @@ class Client:
             not kwargs.get('show_progress'),
             total=len(content) if hasattr(content, '__len__') else None,
         )
+        results = DocumentArray()
         with self._pbar:
             self._client.post(
-                **self._get_post_payload(content, kwargs), on_done=self._gather_result
+                **self._get_post_payload(content, kwargs),
+                on_done=partial(self._gather_result, results=results),
             )
-        return self._unboxed_result
 
-    def _gather_result(self, r):
+        for c in content:
+            if hasattr(c, 'tags') and c.tags.pop('__loaded_by_CAS__', False):
+                c.pop('blob')
+
+        return self._unboxed_result(results)
+
+    def _gather_result(self, response, results: 'DocumentArray'):
         from rich import filesize
 
-        if not self._results:
+        if not results:
             self._pbar.start_task(self._r_task)
-        r = r.data.docs
-        self._results.extend(r)
+        r = response.data.docs
+        results.extend(r)
         self._pbar.update(
             self._r_task,
             advance=len(r),
@@ -130,39 +142,43 @@ class Client:
             ),
         )
 
-    @property
-    def _unboxed_result(self):
-        if self._results.embeddings is None:
+    @staticmethod
+    def _unboxed_result(results: 'DocumentArray'):
+        if results.embeddings is None:
             raise ValueError(
-                'empty embedding returned from the server. '
+                'Empty embedding returned from the server. '
                 'This often due to a mis-config of the server, '
                 'restarting the server or changing the serving port number often solves the problem'
             )
-        return self._results.embeddings if self._return_plain else self._results
+        return (
+            results.embeddings if ('__created_by_CAS__' in results[0].tags) else results
+        )
 
     def _iter_doc(self, content) -> Generator['Document', None, None]:
         from rich import filesize
         from docarray import Document
 
-        self._return_plain = True
-
         if hasattr(self, '_pbar'):
             self._pbar.start_task(self._s_task)
 
         for c in content:
             if isinstance(c, str):
-                self._return_plain = True
                 _mime = mimetypes.guess_type(c)[0]
                 if _mime and _mime.startswith('image'):
-                    yield Document(uri=c).load_uri_to_blob()
+                    yield Document(
+                        tags={'__created_by_CAS__': True, '__loaded_by_CAS__': True},
+                        uri=c,
+                    ).load_uri_to_blob()
                 else:
-                    yield Document(text=c)
+                    yield Document(tags={'__created_by_CAS__': True}, text=c)
             elif isinstance(c, Document):
-                self._return_plain = False
-                if c.content_type in ('text', 'blob', 'tensor'):
+                if c.content_type in ('text', 'blob'):
                     yield c
                 elif not c.blob and c.uri:
                     c.load_uri_to_blob()
+                    c.tags['__loaded_by_CAS__'] = True
+                    yield c
+                elif c.tensor is not None:
                     yield c
                 else:
                     raise TypeError(f'unsupported input type {c!r} {c.content_type}')
@@ -181,18 +197,22 @@ class Client:
                 )
 
     def _get_post_payload(self, content, kwargs):
-        return dict(
-            on='/',
+        parameters = kwargs.get('parameters', {})
+        model_name = parameters.get('model', '')
+        payload = dict(
+            on=f'/encode/{model_name}'.rstrip('/'),
             inputs=self._iter_doc(content),
-            request_size=kwargs.get(
-                'batch_size', 8
-            ),  # the default `batch_size` is very subjective. i would set it 8 based on 2 considerations (1) play safe on most GPUs (2) ease the load to our demo server
+            request_size=kwargs.get('batch_size', 8),
             total_docs=len(content) if hasattr(content, '__len__') else None,
         )
+        if self._scheme == 'grpc' and self._authorization:
+            payload.update(metadata=('authorization', self._authorization))
+        elif self._scheme == 'http' and self._authorization:
+            payload.update(headers={'Authorization': self._authorization})
+        return payload
 
     def profile(self, content: Optional[str] = '') -> Dict[str, float]:
         """Profiling a single query's roundtrip including network and computation latency. Results is summarized in a table.
-
         :param content: the content to be sent for profiling. By default it sends an empty Document
             that helps you understand the network latency.
         :return: the latency report in a dict.
@@ -271,28 +291,40 @@ class Client:
             total=len(content) if hasattr(content, '__len__') else None,
         )
 
-        async for da in self._async_client.post(
-            **self._get_post_payload(content, kwargs)
-        ):
-            if not self._results:
-                self._pbar.start_task(self._r_task)
-            self._results.extend(da)
-            self._pbar.update(
-                self._r_task,
-                advance=len(da),
-                total_size=str(
-                    filesize.decimal(int(os.environ.get('JINA_GRPC_RECV_BYTES', '0')))
-                ),
-            )
+        results = DocumentArray()
+        with self._pbar:
+            async for da in self._async_client.post(
+                **self._get_post_payload(content, kwargs)
+            ):
+                if not results:
+                    self._pbar.start_task(self._r_task)
+                results.extend(da)
+                self._pbar.update(
+                    self._r_task,
+                    advance=len(da),
+                    total_size=str(
+                        filesize.decimal(
+                            int(os.environ.get('JINA_GRPC_RECV_BYTES', '0'))
+                        )
+                    ),
+                )
 
-        return self._unboxed_result
+        for c in content:
+            if hasattr(c, 'tags') and c.tags.pop('__loaded_by_CAS__', False):
+                c.pop('blob')
+
+        return self._unboxed_result(results)
 
     def _prepare_streaming(self, disable, total):
 
         if total is None:
             total = 500
             warnings.warn(
-                'the length of the input is unknown, the progressbar would not be accurate.'
+                'The length of the input is unknown, the progressbar would not be accurate.'
+            )
+        elif total > 500:
+            warnings.warn(
+                'Please ensure all the inputs are valid, otherwise the request will be aborted.'
             )
 
         from docarray.array.mixins.io.pbar import get_pbar
@@ -309,21 +341,18 @@ class Client:
             ':arrow_down: Recv', total=total, total_size=0, start=False
         )
 
-        from docarray import DocumentArray
-
-        self._results = DocumentArray()
-
     @staticmethod
     def _prepare_single_doc(d: 'Document'):
         if d.content_type in ('text', 'blob'):
             return d
         elif not d.blob and d.uri:
             d.load_uri_to_blob()
+            d.tags['__loaded_by_CAS__'] = True
             return d
         elif d.tensor is not None:
             return d
         else:
-            raise TypeError(f'unsupported input type {d!r} {d.content_type}')
+            raise TypeError(f'Unsupported input type {d!r} {d.content_type}')
 
     @staticmethod
     def _prepare_rank_doc(d: 'Document', _source: str = 'matches'):
@@ -334,14 +363,24 @@ class Client:
         setattr(d, _source, [Client._prepare_single_doc(c) for c in _get(d)])
         return d
 
+    @staticmethod
+    def _reset_rank_doc(d: 'Document', _source: str = 'matches'):
+        _get = lambda d: getattr(d, _source)
+
+        if d.tags.pop('__loaded_by_CAS__', False):
+            d.pop('blob')
+
+        for c in _get(d):
+            if c.tags.pop('__loaded_by_CAS__', False):
+                c.pop('blob')
+        return d
+
     def _iter_rank_docs(
         self, content, _source='matches'
     ) -> Generator['Document', None, None]:
         from rich import filesize
         from docarray import Document
 
-        self._return_plain = True
-
         if hasattr(self, '_pbar'):
             self._pbar.start_task(self._s_task)
 
@@ -349,7 +388,7 @@ class Client:
             if isinstance(c, Document):
                 yield self._prepare_rank_doc(c, _source)
             else:
-                raise TypeError(f'unsupported input type {c!r}')
+                raise TypeError(f'Unsupported input type {c!r}')
 
             if hasattr(self, '_pbar'):
                 self._pbar.update(
@@ -363,37 +402,45 @@ class Client:
                 )
 
     def _get_rank_payload(self, content, kwargs):
-        return dict(
-            on='/rank',
+        parameters = kwargs.get('parameters', {})
+        model_name = parameters.get('model', '')
+        payload = dict(
+            on=f'/rank/{model_name}'.rstrip('/'),
             inputs=self._iter_rank_docs(
                 content, _source=kwargs.get('source', 'matches')
             ),
             request_size=kwargs.get('batch_size', 8),
             total_docs=len(content) if hasattr(content, '__len__') else None,
         )
+        if self._scheme == 'grpc' and self._authorization:
+            payload.update(metadata=('authorization', self._authorization))
+        elif self._scheme == 'http' and self._authorization:
+            payload.update(headers={'Authorization': self._authorization})
+        return payload
 
     def rank(self, docs: Iterable['Document'], **kwargs) -> 'DocumentArray':
         """Rank image-text matches according to the server CLIP model.
-
         Given a Document with nested matches, where the root is image/text and the matches is in another modality, i.e.
         text/image; this method ranks the matches according to the CLIP model.
-
         Each match now has a new score inside ``clip_score`` and matches are sorted descendingly according to this score.
         More details can be found in: https://github.com/openai/CLIP#usage
-
         :param docs: the input Documents
         :return: the ranked Documents in a DocumentArray.
-
         """
         self._prepare_streaming(
             not kwargs.get('show_progress'),
             total=len(docs),
         )
+        results = DocumentArray()
         with self._pbar:
             self._client.post(
-                **self._get_rank_payload(docs, kwargs), on_done=self._gather_result
+                **self._get_rank_payload(docs, kwargs),
+                on_done=partial(self._gather_result, results=results),
             )
-        return self._results
+        for d in docs:
+            self._reset_rank_doc(d, _source=kwargs.get('source', 'matches'))
+
+        return results
 
     async def arank(self, docs: Iterable['Document'], **kwargs) -> 'DocumentArray':
         from rich import filesize
@@ -402,17 +449,25 @@ class Client:
             not kwargs.get('show_progress'),
             total=len(docs),
         )
+        results = DocumentArray()
+        with self._pbar:
+            async for da in self._async_client.post(
+                **self._get_rank_payload(docs, kwargs)
+            ):
+                if not results:
+                    self._pbar.start_task(self._r_task)
+                results.extend(da)
+                self._pbar.update(
+                    self._r_task,
+                    advance=len(da),
+                    total_size=str(
+                        filesize.decimal(
+                            int(os.environ.get('JINA_GRPC_RECV_BYTES', '0'))
+                        )
+                    ),
+                )
 
-        async for da in self._async_client.post(**self._get_rank_payload(docs, kwargs)):
-            if not self._results:
-                self._pbar.start_task(self._r_task)
-            self._results.extend(da)
-            self._pbar.update(
-                self._r_task,
-                advance=len(da),
-                total_size=str(
-                    filesize.decimal(int(os.environ.get('JINA_GRPC_RECV_BYTES', '0')))
-                ),
-            )
+        for d in docs:
+            self._reset_rank_doc(d, _source=kwargs.get('source', 'matches'))
 
-        return self._results
+        return results
